/***********************************************************
 * $Id: BeanImporter.java 223 2010-11-25 10:01:20Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.server.UID;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.RemoteInvocation;
import org.clazzes.remoting.RemoteInvocationFailureException;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;

/**
 * Imports a service exported by {@link BeanInvocationHandler} via remoting.
 */
public class BeanImporter extends ReturnCallbackSupport {

	private static final Log log = LogFactory.getLog(BeanImporter.class);
		
	private ClientManager clientManager;
	private Class<?> serviceInterface;
	private ClassLoader proxyClassLoader;
	private String subSystem;
	private Object proxy;
	private List<CallbackWrapper> callbackWrappers;
	private boolean reconnectOnTransportError;
    private boolean bindClientToReturnCallbacks;
    private String returnCallbackSubSystem;
    	
	private static class CallbackArgument {
		
		public final int position;
		public final CallbackWrapper callbackHandler;
		
		public CallbackArgument(int position,
				CallbackWrapper callbackHandler) {
			super();
			this.position = position;
			this.callbackHandler = callbackHandler;
		}
	};
	
	private Map<Method,List<CallbackArgument>> callbackMethods;

    private class ReturnInvocationHandler implements InvocationHandler {

        private final Map<String,Object> payload;
        private final Object proxyGUID;
        private final Map<String, MethodInfo> methodInfos;
        private boolean closed;
	    
        /**
         * Construct a new return callback handler.
         * 
         * @param proxyGUID The GUID for identifying the callback invocation handler on the server side.
         * @param methodInfos If not <code>null</code>, this instance must call
         *                    {@link ClientHolder#releaseFromReturnCallback(ClientManager)}, when
         *                    a close method is invoked. Additionally, duplicate calls to close
         *                    methods will be silently ignored. 
         */
        public ReturnInvocationHandler(Object proxyGUID, Map<String, MethodInfo> methodInfos) {
	            
            this.proxyGUID = proxyGUID;
            this.payload = new HashMap<String, Object>(3); 
            this.methodInfos = methodInfos;
	                
            this.payload.put(ReturnCallbackHandler.RETURN_PROXY_GUID_PROPERTY,proxyGUID);
        }

        /* (non-Javadoc)
         * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
         */
        public Object invoke(Object proxy, Method method, Object[] args)
                            throws Throwable {
	            
            Client client = ClientHolder.getContext(BeanImporter.this.clientManager);
	            
            boolean threadBoundClient;
            MethodInfo mi = null;
            
            if (this.methodInfos != null) {
                
                mi = this.methodInfos.get(method.getName());
                
                // ignore duplicate calls to close.
                // However, do not ignore calls to ordinary methods after close().
                // The rationale for this is to not introduce too much logic to the client.
                // Calls to a closed callback proxy will crash at the server side
                // with a reported RemoteInvocationFailure exception anyways.
                if (mi != null && mi.closeMethod && this.closed) {
                    
                    if (log.isDebugEnabled())
                        log.debug("Ignoring duplicate call to close method ["+method.getName()+"] on subsystem ["+BeanImporter.this.returnCallbackSubSystem+"], GUID ["+this.proxyGUID+"]...");
                  
                    return null;
                }
            }

            if (client == null) {
                threadBoundClient = false;
                client = BeanImporter.this.clientManager.activateClient();
            }
            else {
                threadBoundClient = true;
            }
	        
            try {
                
                Object res;
	                
                try {
                    
                    if (log.isDebugEnabled())
                        log.debug("Calling method ["+method.getName()+"] on subsystem ["+BeanImporter.this.returnCallbackSubSystem+"], GUID ["+this.proxyGUID+"]...");
                    
                    res = client.invoke(BeanImporter.this.returnCallbackSubSystem,new RemoteInvocation(method.getName(),args),this.payload);
                    
                    if (log.isDebugEnabled())
                        log.debug("Successfully called method ["+method.getName()+"] on subsystem ["+BeanImporter.this.returnCallbackSubSystem+"], GUID ["+this.proxyGUID+"].");
                    
                } catch (IOException t) {
                    
                    log.error("Disconnecting client upon transport error.",t);
                    client.disconnect();
                    
                    throw t;
                }
                
                catch (InvocationTargetException ite) {
                    
                    // combine the foreign and the local stack trace.
                    Throwable e = ite.getCause();
                    
                    StackTraceElement[] foreignStack = e.getStackTrace();
                    e.fillInStackTrace();
                    StackTraceElement[] myStack = e.getStackTrace();
                    StackTraceElement[] combinedStack = new StackTraceElement[foreignStack.length+myStack.length];
                    System.arraycopy(foreignStack,0,combinedStack,0,foreignStack.length);
                    System.arraycopy(myStack,0,combinedStack,foreignStack.length,myStack.length);
                    e.setStackTrace(combinedStack);
                    throw e;            
                }
                
                return res;
                
            } finally {
                if (!threadBoundClient) {
                    BeanImporter.this.clientManager.passivateClient(client);
                }
                else if (mi != null && mi.closeMethod && !this.closed) {
                    
                    ClientHolder.releaseFromReturnCallback(BeanImporter.this.clientManager);
                    this.closed = true;
                }
            }
        }
    }
    
	private class ClientInvocationHandler implements InvocationHandler {

	    private Object invokeInner(Client client, Method method, Object []args) throws InvocationTargetException, IOException {
	        
            try {
                return client.invoke(BeanImporter.this.subSystem,new RemoteInvocation(method.getName(),args),null);
                
            } catch (IOException e) {
                    
                if (BeanImporter.this.reconnectOnTransportError) {
                    log.warn("Transport error while invoking method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"], trying to reconnect.",e);
                    client.disconnect();
                    client.connect();
                    log.info("Successfully reconnected client.");
                    return client.invoke(BeanImporter.this.subSystem,new RemoteInvocation(method.getName(),args),null);
                }
                else
                    throw e;
            }
	    }
	    
	    /* (non-Javadoc)
		 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
		 */
		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
              
		    Client client = ClientHolder.bindToReturnCallback(BeanImporter.this.clientManager);
            			
			try {
				Object res;
				
				try {
					
					if (log.isDebugEnabled())
						log.debug("Calling method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"]...");
					
					List<CallbackArgument> callbackArguments = null;
					
					if (BeanImporter.this.callbackMethods!=null)
						callbackArguments = BeanImporter.this.callbackMethods.get(method);
					
					if (callbackArguments == null) {
					    
					    res = this.invokeInner(client,method,args);
					}
					else {
						
						Object[] newArgs  = new Object[args.length];
						
						int ilast = 0;
						
						UID[] cbIds = new UID[callbackArguments.size()];
						
						int nCallbacks = 0;
						
						for (CallbackArgument ca : callbackArguments) {
							
							Object arg = args[ca.position];
							
							if (arg != null) {
								
								if (ca.position > ilast)
									System.arraycopy(args,ilast,newArgs,ilast,ca.position-ilast);
								
								cbIds[nCallbacks] = client.registerCallback(
										ca.callbackHandler.getWrapper(args[ca.position]));
							
								if (log.isDebugEnabled())
									log.debug("Registered callback no. ["+nCallbacks+"] for argument ["+ca.position+"] with UID ["+cbIds[nCallbacks]+"]...");

								newArgs[ca.position] = cbIds[nCallbacks];
				
								++nCallbacks;
								ilast = ca.position + 1;
							}
						}
						
						if (ilast < args.length)
							System.arraycopy(args,ilast,newArgs,ilast,args.length-ilast);
						
						try {
						    res = this.invokeInner(client,method,newArgs);			                
						} finally {
							
							for (int icb = 0; icb<nCallbacks; ++icb) {
								
								CallbackArgument ca =  callbackArguments.get(icb);
							
								if (log.isDebugEnabled())
									log.debug("deregistered callback for argument ["+ca.position+"] with GUID ["+cbIds[icb]+"]...");
								
								client.deregisterCallback(cbIds[icb]);
							}
						}
					}
					
					if (log.isDebugEnabled())
						log.debug("Successfully called method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"].");

				} catch (IOException t) {
					
				    log.error("Disconnecting client upon transport error.",t);
				    client.disconnect();
					
					throw new RemoteInvocationFailureException("Call to method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"] failed with transport error.",t);
				}
				catch (InvocationTargetException ite)  {
					
					// combine the foreign and the local stack trace.
					Throwable e = ite.getCause();
					
					StackTraceElement[] foreignStack = e.getStackTrace();
					e.fillInStackTrace();
					StackTraceElement[] myStack = e.getStackTrace();
					StackTraceElement[] combinedStack = new StackTraceElement[foreignStack.length+myStack.length];
					System.arraycopy(foreignStack,0,combinedStack,0,foreignStack.length);
					System.arraycopy(myStack,0,combinedStack,foreignStack.length,myStack.length);
					e.setStackTrace(combinedStack);
					throw e;			
				}
				
				if (res != null) {
				    
		            Class<?> rt = method.getReturnType();
		            Map<String, MethodInfo> rcbInfo = BeanImporter.this.getCallbackInterfaceMethods(rt);

		            if (rcbInfo != null) {

			            if (log.isDebugEnabled())
			                log.debug("Creating proxy for interface ["+rt.getName()+"], GUID ["+res+"].");
			                
			            ClassLoader cl = BeanImporter.this.proxyClassLoader;
			            
			            if (cl == null)
			                cl = rt.getClassLoader();
			            
			            // keep client bound to current thread until a close method
			            // of the returned callback is invoked.
			            if (BeanImporter.this.bindClientToReturnCallbacks)
			                client = null;

			            return
			            Proxy.newProxyInstance(cl,
			                    new Class<?>[] {rt},new ReturnInvocationHandler(res, BeanImporter.this.bindClientToReturnCallbacks ? rcbInfo : null));
			        }
				}
				
				return res;
				
			} finally {
				if (client != null)
					ClientHolder.releaseFromReturnCallback(BeanImporter.this.clientManager);
			}
		}
	}
	
	/* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#isInitialized()
     */
    public boolean isInitialized() {
        
        return this.proxy != null;
    }

    /* (non-Javadoc)
	 * @see org.clazzes.remoting.Initializable#initialize()
	 */
	public void initialize() throws Exception {
		
	    super.initialize();
	    
		if (this.callbackWrappers != null) {
			
			this.callbackMethods = new HashMap<Method, List<CallbackArgument>>();
		
			for (Method method : this.serviceInterface.getMethods()) {
			
				List<CallbackArgument> callbackArguments = null;
			
				for (int iarg = 0; iarg < method.getParameterTypes().length; ++iarg){
					
					Class<?> arg = method.getParameterTypes()[iarg];
					
					for (CallbackWrapper handler : this.callbackWrappers) {
					
						if (handler.getServiceInterface().isAssignableFrom(arg))
						{
							if (callbackArguments == null)
								callbackArguments = new ArrayList<CallbackArgument>();
							
							if (log.isDebugEnabled())
								log.debug("Argument ["+iarg+"] of method ["+method+"] will be treated as callback with interface ["+handler.getServiceInterface().getName()+"].");
							
							callbackArguments.add(new CallbackArgument(iarg,handler));
						}
					}
				}
				
				if (callbackArguments != null)
					this.callbackMethods.put(method,callbackArguments);
			}
		}
		
		ClassLoader cl = this.proxyClassLoader;
		
		if (cl == null)
		    cl = this.serviceInterface.getClassLoader();
		
		this.proxy=
			Proxy.newProxyInstance(cl,
					new Class<?>[] {this.serviceInterface},new ClientInvocationHandler());		
	}

	/**
	 * @return The proxy, which has been set up in {@link #initialize()}.
	 */
	public Object getProxy() {
		
		return this.proxy;
	}

	/**
	 * @return the clientManager
	 */
	public ClientManager getClientManager() {
		return this.clientManager;
	}

	/**
	 * @param clientManager the clientManager to set
	 */
	public void setClientManager(ClientManager clientManager) {
		this.clientManager = clientManager;
	}

	/**
	 * @return the jboss remoting subsystem.
	 * @see Client#setSubsystem(String)
	 */
	public String getSubSystem() {
		return this.subSystem;
	}

	/**
	 * @param subSystem the subSystem to set
	 */
	public void setSubSystem(String subSystem) {
		this.subSystem = subSystem;
	}

	/**
     * @return the sub system used to call for returned callback objects. Set this property,
     *         if you also called {@link #setReturnCallbackInterfaces(List)} in order to specify
     *         return callbacks.
     */
    public String getReturnCallbackSubSystem() {
        return this.returnCallbackSubSystem;
    }

    /**
     * @param returnCallbackSubSystem the returnCallbackSubSystem to set
     */
    public void setReturnCallbackSubSystem(String returnCallbackSubSystem) {
        this.returnCallbackSubSystem = returnCallbackSubSystem;
    }

    /**
	 * @return the service interface implemented by the proxy.
	 */
	public Class<?> getServiceInterface() {
		return this.serviceInterface;
	}

	/**
	 * @param serviceInterface the serviceInterface to set.
	 */
	public void setServiceInterface(Class<?> serviceInterface) {
		this.serviceInterface = serviceInterface;
	}

	/**
     * @return the class loader to pass to {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)}.
     *         If <code>null</code>, <code>serviceInterface.class.getClassLoader()</code> is used
     *         for defining proxy classes.
     * 
     */
    public ClassLoader getProxyClassLoader() {
        return this.proxyClassLoader;
    }

    /**
     * @param proxyClassLoader the class loader to pass to
     *         {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)}.
     *         If set to the default value <code>null</code>,
     *         <code>serviceInterface.class.getClassLoader()</code> is used
     *         for defining proxy classes.
     */
    public void setProxyClassLoader(ClassLoader proxyClassLoader) {
        this.proxyClassLoader = proxyClassLoader;
    }

    /**
	 * @return the callbackHandlers
	 */
	public List<CallbackWrapper> getCallbackWrappers() {
		return this.callbackWrappers;
	}

	/**
	 * @param callbackHandlers the callbackHandlers to set
	 */
	public void setCallbackWrappers(List<CallbackWrapper> callbackHandlers) {
		this.callbackWrappers = callbackHandlers;
	}

    /**
     * @return Whether to bind clients to the current thread for the lifetime of a returned callback.
     */
    public boolean isBindClientToReturnCallbacks() {
        return this.bindClientToReturnCallbacks;
    }

    /**
     * @param bindClientToReturnCallbacks Whether to bind clients to the current thread for
     *                                    the lifetime of a returned callback.
     */
    public void setBindClientToReturnCallbacks(boolean bindClientToReturnCallbacks) {
        this.bindClientToReturnCallbacks = bindClientToReturnCallbacks;
    }

    /**
     * @return Whether we reconnect on transport errors.
     */
    public boolean isReconnectOnTransportError() {
        return this.reconnectOnTransportError;
    }

    /**
     * @param reconnectOnTransportError Whether we reconnect on transport errors.
     *                  Reconnecting on transport errors is handled by
     *                  {@link OpenClientContextInViewFilter and {@link OpenClientSessionInterceptor},
     *                  so setting this flag to <code>true</code> is mostly an option,
     *                  if you do not use session contexts. When setting this option
     *                  to <code>true</code>, you risk, that your remote business logic
     *                  is called twice, when the transport error occurs when transmitting
     *                  the reponse from the server to the client.
     */
    public void setReconnectOnTransportError(boolean reconnectOnTransportError) {
        this.reconnectOnTransportError = reconnectOnTransportError;
    }
}
