/***********************************************************
 * $Id: ClientHolder.java 230 2011-02-05 16:02:31Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.TimeoutException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;
import org.clazzes.util.aop.ThreadLocalManager;

/**
 * This class provides static helper functions for thread-bound session contexts, which
 * may be used by session interceptors or servlet filters.
 */
public class ClientHolder {

    private static final Log log = LogFactory.getLog(ClientHolder.class);
    
    private final Client client;
    private boolean hasSession;
    private int nCallbacks;
    
    ClientHolder(final Client client) {
        this.nCallbacks = 0;
        this.client = client;
        this.hasSession = false;
    }
    
    void addCallback() {
        ++this.nCallbacks;
    }
    
    void removeCallback() {
        --this.nCallbacks;
    }
    
    Client getClient() {
        return this.client;
    }
    
    void setHasSession(boolean hasSession) {
        this.hasSession = hasSession;
    }
    
    boolean isHasSession() {
        return this.hasSession;
    }
    
    boolean isDisposable() {
        return !this.client.isConnected() || (!this.hasSession && this.nCallbacks == 0);
    }
    
    static ClientHolder getContextInner(ClientManager clientManager) {
        
        return ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
    }

    /**
     * Get a thread-bound client for a client manager, which has previously been opened with
     * {@link #openContext(ClientManager, String, SessionAuthenticator)}.
     * 
     * @param clientManager The client manager for which to open the session.
     * @return The client holder attached to the current thread or <code>null</code>,
     *         if no context is bound to the current thread. 
     */
    public static Client getContext(ClientManager clientManager) {
        
        ClientHolder ch = ClientHolder.getContextInner(clientManager);
        if (ch==null) return null;
        return ch.getClient();
    }
    
    private static void openSession(Client client,
            String sessionContextSubSystem,
            SessionAuthenticator sessionAuthenticator) throws InvocationTargetException, IOException {

        if (sessionAuthenticator == null) {
            
            client.invoke(sessionContextSubSystem,SessionCmd.OPEN_SESSION,null);
        }
        else {
            
            client.invoke(sessionContextSubSystem,SessionCmd.OPEN_SESSION,
                    sessionAuthenticator.createCredentials(client.getClientUID()));
        }

    }
    
    private static void closeSession(Client client, String sessionContextSubSystem) throws InvocationTargetException, IOException {
        client.invoke(sessionContextSubSystem,SessionCmd.CLOSE_SESSION,null);
    }
  
    /**
     * Open a thread-bound session context for the given client manager.
     * 
     * @param clientManager The client manager for which to open the session.
     * @param sessionContextSubSystem The subsystem name of the remote session context subsystem.
     * @param sessionAuthenticator An optional session authenticator, which may be null.
     * @return The new client holder.
     * @throws TimeoutException If no more clients are available.
     * @throws IOException If an unrecoverable transport error occurred.
     * @throws InvocationTargetException If the remote system threw an exception upon opening the sessoin context.
     */
    public static Client openContext(ClientManager clientManager, 
            String sessionContextSubSystem,
            SessionAuthenticator sessionAuthenticator) throws TimeoutException, IOException, InvocationTargetException {
        
        ClientHolder client = ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
        
        if (client != null && client.isHasSession()) {
            
            log.warn("Closing left-over session for current thread ["+Thread.currentThread().getName()+"] before trying to open session.");
            closeContextInner(client,clientManager,sessionContextSubSystem);
        }
        
        client = new ClientHolder(clientManager.activateClient());
        
        if (log.isDebugEnabled())
            log.debug("Binding client session context to current thread ["+Thread.currentThread().getName()+"].");

        try {
            
            if (log.isDebugEnabled())
                log.debug("Opening remote session for current thread ["+Thread.currentThread().getName()+"].");         

            try {
                openSession(client.getClient(), sessionContextSubSystem, sessionAuthenticator);
            } catch (IOException e) {
                log.warn("Transport error while opening session, trying to reconnect.",e);
                client.getClient().disconnect();
                client.getClient().connect();
                openSession(client.getClient(), sessionContextSubSystem, sessionAuthenticator);
                log.info("Successfully reconnected client.");
            }
            
            client.setHasSession(true);
        }
        catch (Throwable e) {
            client.getClient().disconnect();
            clientManager.passivateClient(client.getClient());
        }
        
        ThreadLocalManager.bindResource(clientManager.getUniqueID(),client);
        return client.getClient();
    }
    
    /**
     * Close a thread-bound remote session context for the given client manager.
     * 
     * @param clientManager The client manager for which to close the remote session.
     * @param sessionContextSubSystem The subsystem name of the remote session context subsystem.
     */
    public static void closeContext(ClientManager clientManager, 
            String sessionContextSubSystem) {
        
        ClientHolder client = getContextInner(clientManager);
        
        if (client == null) {
            
            log.warn("No session context for current thread ["+Thread.currentThread().getName()+"] when trying to close session.");
            return;
        }
        
        closeContextInner(client,clientManager,sessionContextSubSystem);
    }
    
    private static void closeContextInner(ClientHolder client,
            ClientManager clientManager, 
            String sessionContextSubSystem) {
            
        if (client.getClient().isConnected()) {
            try {
                if (log.isDebugEnabled())
                    log.debug("Closing remote session of current thread ["+Thread.currentThread().getName()+"].");           
                
                closeSession(client.getClient(),sessionContextSubSystem);
            }
            catch (Throwable e) {
                log.warn("Error closing remote session of current thread ["+Thread.currentThread().getName()+"].",e);
            }
            client.setHasSession(false);
        }
        
        if (client.isDisposable()) {
            clientManager.passivateClient(client.getClient());
            ThreadLocalManager.unbindResource(clientManager.getUniqueID());
        }
        
        if (log.isDebugEnabled())
            log.debug("Unbinding client session context Session from current thread ["+Thread.currentThread().getName()+"].");          
    }
    
    /**
     * Bind a client to a thread for the lifetime of a return callback.
     * 
     * @param clientManager The client manager used to retrieve a thread-bound client or
     *                      to fetch a new session.
     * @throws TimeoutException Upon errors fetching a new client from a client manager.
     */
    public static Client bindToReturnCallback(ClientManager clientManager) throws TimeoutException {
        
        ClientHolder client = ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
        
        if (client == null) {    
            client = new ClientHolder(clientManager.activateClient());
        }
        
        client.addCallback();
        
        if (log.isDebugEnabled())
            log.debug("Binding context client to current thread ["+Thread.currentThread().getName()+"].");

        ThreadLocalManager.bindResource(clientManager.getUniqueID(),client);
        return client.getClient();
    }
    
    /**
     * Release a thread-bind session previously bound by {@link #bindToReturnCallback(ClientManager)}
     * from a callback.
     * 
     * @param clientManager The client manager used to retrieve a thread-bound client 
     */
    public static void releaseFromReturnCallback(ClientManager clientManager) {

        ClientHolder client = ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
        
        if (client == null) {
            
            log.warn("No context client for current thread ["+Thread.currentThread().getName()+"] when trying to release a callback-nound session.");
            return;
        }

        client.removeCallback();
        
        if (client.isDisposable()) {
            clientManager.passivateClient(client.getClient());
            ThreadLocalManager.unbindResource(clientManager.getUniqueID());
        }
        
        if (log.isDebugEnabled())
            log.debug("Unbinding context client from current thread ["+Thread.currentThread().getName()+"].");          
    }
}
