/***********************************************************
 * $Id: ReturnCallbackSupport.java 219 2010-11-23 12:29:05Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 23.11.2010
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.InvalidClassException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Initializable;
import org.clazzes.remoting.annotations.CloseMethod;

/**
 * A support class for all classes, which implement return callback functionality and need
 * knowledge about callback interfaces and close mathods in these interfaces.
 */
public abstract class ReturnCallbackSupport implements Initializable {
    
    private static final Log log = LogFactory.getLog(ReturnCallbackSupport.class);
    
    protected static class MethodInfo {
        
        public final Method method;
        public final boolean closeMethod;
        
        public MethodInfo(Method method, boolean closeMethod) {
            super();
            this.method = method;
            this.closeMethod = closeMethod;
        }
    }

    private Map<Class<?>,Map<String,MethodInfo>> callbackServiceMethods;
    private Map<Class<?>,List<Method>> callbackCloseMethods;
    private List<Class<?>> returnCallbackInterfaces;
        
    private String closeMethodRegex;

    /* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#initialize()
     */
    public void initialize() throws Exception {

        this.callbackServiceMethods = new HashMap<Class<?>, Map<String,MethodInfo>>();
        this.callbackCloseMethods = new HashMap<Class<?>, List<Method>>();
        
        Pattern closePattern =
            this.closeMethodRegex == null ? null : Pattern.compile(this.closeMethodRegex);
        
        for (Class<?> serviceInterface : this.returnCallbackInterfaces) {
        
            Method [] methods = serviceInterface.getMethods();
        
            Map<String, MethodInfo> serviceMethods = new HashMap<String, MethodInfo>();
            List<Method> closeMethods = null;
            
            for (Method method : methods) {
            
                MethodInfo om = serviceMethods.get(method.getName());
                
                if (om != null)
                    throw new InvalidClassException("Method ["+om.method+"] is duplicated in interface ["+serviceInterface.getName()+"] (duplicate is ["+method+"]).");
                
                boolean closeMethod = closePattern == null ? false :
                    closePattern.matcher(method.getName()).matches();
                
                // check for annotation configuration.
                if (!closeMethod && method.getAnnotation(CloseMethod.class) != null)
                    closeMethod = true;
                
                if (log.isDebugEnabled()) {
                    if (closeMethod)
                        log.debug("Adding close method ["+ method +"] to the list of service methods for interface ["+serviceInterface.getName()+"].");
                    else
                        log.debug("Adding method ["+ method +"] to the list of service methods for interface ["+serviceInterface.getName()+"].");
                }
                
                serviceMethods.put(method.getName(),new MethodInfo(method,closeMethod));
                
                if (closeMethod) {    
                    if (closeMethods == null) closeMethods = new LinkedList<Method>();
                    closeMethods.add(method);
                }
            }
            this.callbackServiceMethods.put(serviceInterface,serviceMethods);
            
            if (closeMethods != null)
                this.callbackCloseMethods.put(serviceInterface,closeMethods);
        }

    }

    protected MethodInfo getMethodInfo(Class<?> iface, String methodName) {
        
        Map<String, MethodInfo> ifInfo = this.callbackServiceMethods.get(iface);
        if (ifInfo == null) return null;
        return ifInfo.get(methodName);
    }
    
    protected List<Method> getCloseMethods(Class<?> iface) {
        return this.callbackCloseMethods.get(iface);
    }

    protected Map<String, MethodInfo> getCallbackInterfaceMethods(Class<?> iface) {
        
        return this.callbackServiceMethods.get(iface);
    }

    /**
     * @return the callbackInterfaces
     */
    public List<Class<?>> getReturnCallbackInterfaces() {
        return this.returnCallbackInterfaces;
    }

    /**
     * @param callbackInterfaces the callbackInterfaces to set
     */
    public void setReturnCallbackInterfaces(List<Class<?>> callbackInterfaces) {
        this.returnCallbackInterfaces = callbackInterfaces;
    }

    /**
     * @return The regular expression, which is used to identify context lose methods.
     *         A context close method deregisters the callback object and hence marks the end
     *         of the lifecycle of a return proxy.
     */
    public String getCloseMethodRegex() {
        return this.closeMethodRegex;
    }

    /**
     * @param closeMethodRegex the closeMethodRegex to set
     */
    public void setCloseMethodRegex(String closeMethodRegex) {
        this.closeMethodRegex = closeMethodRegex;
    }

}
