/***********************************************************
 * $Id: ClientParameters.java 54 2009-04-16 16:45:08Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 15.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.client;

import java.io.IOException;
import java.net.Socket;
import java.util.Map;

import javax.net.SocketFactory;

import org.clazzes.remoting.SocketParameters;


/**
 * An object holding the configuration options for a client.
 * 
 * @author wglas
 */
public class ClientParameters extends SocketParameters {

    protected SocketFactory socketFactory;
    int soTimeout;
    boolean asynchronous;

    /**
     * Construct a client parameters object with the following default parameters:
     * <table>
     *   <tr><td>soTimeout</td><td>60000</td></tr>
     *   <tr><td>tcpNoDelay</td><td>true</td></tr>
     *   <tr><td>asynchronous</td><td>false</td></tr>
     * </table>
     */
    public ClientParameters() {
        super();
        
        this.soTimeout = 60000;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.SocketParameters#tuneSocket(java.net.Socket)
     */
    @Override
    public Map<String, Object> tuneSocket(Socket socket) throws IOException {
        
        Map<String, Object> ret = super.tuneSocket(socket);
        socket.setSoTimeout(this.soTimeout);
        return ret;
    }

    /**
     * @return The socket factory used by this client. If not set by
     *    {@link #setSocketFactory(SocketFactory)}, {@link SocketFactory#getDefault()}
     *    is returned.
     */
    public SocketFactory getSocketFactory() {
        
        if (this.socketFactory == null)
            this.socketFactory = SocketFactory.getDefault();
        
        return this.socketFactory;
    }

    /**
     * @param socketFactory A custom socket factory to set.
     */
    public void setSocketFactory(SocketFactory socketFactory) {
        
        this.socketFactory = socketFactory;
    }

    /**
     * @return The socket timeout used when establishing new connections.
     * 
     * @see Socket#getSoTimeout()
     */
    public int getSoTimeout() {
        return this.soTimeout;
    }

    /**
     * @param soTimeout The socket timeout during the connection establishment phase to set.
     * 
     * @see Socket#setSoTimeout(int)
    */
    public void setSoTimeout(int soTimeout) {
        this.soTimeout = soTimeout;
    }

    /**
     * @return Whether the client should use a dedicated thread for reading responses.
     *         If set to <code>true</code>, the client may handle broadcast and callback
     *         request at any time and not only when waiting for a response to a remote
     *         invocation. If set to <code>false</code>, the invocation request is guaranteed
     *         to be deserialized by the same thread that will serialized the invocation
     *         response afterwards. 
     */
    public boolean isAsynchronous() {
        return this.asynchronous;
    }

    /**
     * @param asynchronous Whether the client should use a dedicated thread for reading responses.
     */
    public void setAsynchronous(boolean asynchronous) {
        this.asynchronous = asynchronous;
    }
}