/***********************************************************
 * $Id: ThreadedClientConnectionHandler.java 54 2009-04-16 16:45:08Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 13.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.client;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.rmi.server.UID;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.InvalidTargetException;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.OneWayHandler;
import org.clazzes.remoting.RemoteInvoker;
import org.clazzes.remoting.cmd.CallBackCmd;
import org.clazzes.remoting.cmd.Cmd;
import org.clazzes.remoting.cmd.CmdVisitorSupport;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.cmd.InvocationExceptionCmd;
import org.clazzes.remoting.cmd.InvocationResultCmd;
import org.clazzes.remoting.cmd.LoadClassCmd;
import org.clazzes.remoting.cmd.ReturnClassBytesCmd;
import org.clazzes.remoting.loading.ClassBytes;
import org.clazzes.remoting.marshal.Marshaler;

/**
 * This is a simple, single-threaded connection handler, which allows callbacks
 * and broadcasts only during the execution of a remote call.
 * 
 * @author wglas
 */
class ThreadedClientConnectionHandler extends CmdVisitorSupport implements ClientConnectionHandler, Runnable {

    private static final Log log = LogFactory.getLog(ThreadedClientConnectionHandler.class);

    private final Marshaler marshaler;
    public final UID clientUID;
    public final RemoteInvoker client;
    private final Map<String, Object> meta;
    
    // connection state.
    private InvocationCmd invocation;
    private InvocationResultCmd invocationResult;
    private InvocationExceptionCmd invocationException;
    private LoadClassCmd loadClass;
    private ReturnClassBytesCmd returnClassBytes;

    private boolean running;

    private Throwable exception;
    
    private Map<String, Object> buildMetaData(Map<String, Object> m) {
        
        if (m == null)
            return this.meta;
        
        if (this.meta != null)
            m.putAll(this.meta);
        
        return m;
    }

    public ThreadedClientConnectionHandler(final Marshaler marshaler, final UID clientUID,
                                           final RemoteInvoker client, final Map<String, Object> meta) {
        
        this.marshaler = marshaler;
        this.clientUID = clientUID;
        this.client = client;
        this.meta = meta;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientConnectionHandler#getClientUID()
     */
    public UID getClientUID() {
        return this.clientUID;
    }

    public synchronized boolean isRunning() {
        return this.running;
    }
    
    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {

        synchronized(this) {
            this.running = true;
        }
            
        if (log.isDebugEnabled())
            log.debug("Reader thread for client ["+this.clientUID+"] has started.");

        while (this.isRunning()) {

            try {
                Object obj = this.marshaler.readObject();
                
                Cmd.acceptCmd(obj,this);
                
            }
            catch(Throwable e) {
                
                synchronized(this) {
                    
                    this.exception = e;
                    this.notifyAll();
                }
            }
        }

        if (log.isDebugEnabled())
            log.debug("Reader thread for client ["+this.clientUID+"] has finished.");
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientConnectionHandler#close()
     */
    public synchronized void close() {
        
        this.marshaler.close();
        this.running = false;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.cmd.CmdVisitorSupport#visit(org.clazzes.remoting.cmd.InvocationCmd)
     */
    @Override
    public void visit(InvocationCmd cmd) {
       
        // handle broadcast from server.
        String subSystem = cmd.getSubSystem();
        
        OneWayHandler handler = this.client.getBroadcastHandler(subSystem);
        
        if (handler == null) {
            
            log.warn("No broadcast handler for subsystem ["+subSystem+"] registered.");
            
        } else {
            
            try {
                
                if (log.isDebugEnabled())
                    log.debug("Invoking broadcast handler for subsystem ["+subSystem+"]...");
            
                handler.doOneWay(this.client,cmd.getArg(),
                                 this.buildMetaData(cmd.getMetadata()));

                if (log.isDebugEnabled())
                    log.debug("Successfully invoked broadcast handler for subsystem ["+subSystem+"].");

            } catch (Throwable e) {
                log.warn("Broadcast handler for subsystem ["+subSystem+"] failed",e);
            }
        }
    }

    @Override
    public void visit(CallBackCmd cmd) {
       
        UID uid = cmd.getCallbackUID();
        
        InvocationHandler handler = this.client.getCallback(uid);
        
        Cmd response;
        
        if (handler == null) {
            
          response=
              new InvocationExceptionCmd(
                      new InvalidTargetException("Callback UID ["+uid+"] does not exist or is expired."));
        
        } else {
            
            try {
                Object res = handler.invoke(this.client,cmd.getArg(),
                                            this.buildMetaData(cmd.getMetadata()));
                
                response = new InvocationResultCmd(res);
                
            } catch (Throwable e) {
                response = new InvocationExceptionCmd(e);
            }
        }
        
        try {
            this.marshaler.writeObject(response);
            
        } catch (IOException e) {
            throw new RuntimeException("I/O exception writing callback response.",e);
        }
    }

    @Override
    public synchronized void visit(InvocationExceptionCmd cmd) {
        
        if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
            super.visit(cmd);
            return;
        }
        
        this.invocationException = cmd;
        this.notifyAll();
    }

    @Override
    public synchronized void visit(InvocationResultCmd cmd) {
        
        if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
            super.visit(cmd);
            return;
        }
            
        this.invocationResult = cmd;
        this.notifyAll();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientConnectionHandler#send(org.clazzes.remoting.cmd.InvocationCmd)
     */
    public synchronized Object send(InvocationCmd invocation) throws IOException, InvocationTargetException {
        
        if (this.invocation != null || this.loadClass != null)
            throw new IOException("Concurrent remote invocations are not implemented.");
                  
        this.marshaler.writeObject(invocation);
        this.invocation = invocation;
        
        try {

            this.wait();

        } catch (InterruptedException e) {
           
            this.invocation = null;
            this.marshaler.close();
            throw new IOException("Wait for invocation result has been interupted.");
        } 
                    
        this.invocation = null;
 
        if  (this.invocationResult != null) {
                
            Object ret = this.invocationResult.getResult();
            this.invocationResult = null;
            return ret;
        }
        
        if  (this.invocationException != null) {
            
            Throwable e = this.invocationException.getException();
            this.invocationException = null;
            throw new InvocationTargetException(e);
        }
      
        if (this.exception instanceof IOException) {
            
            Throwable e = this.exception;
            this.exception = null;
            throw (IOException)e;
        }
            
        if (this.exception != null) {
            IOException ioe = new IOException("Exception caught during read of response to remote invocation.");
            ioe.initCause(this.exception);
            this.exception = null;
            throw ioe;
        }
            
            
        throw new IllegalStateException("Illegal state while fetching response to remote invocation.");
    }
    
    @Override
    public synchronized void visit(ReturnClassBytesCmd cmd) {
        
        if (this.loadClass == null || this.returnClassBytes != null) {
            super.visit(cmd);
            return;
        }
        this.returnClassBytes = cmd;
        this.notifyAll();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientConnectionHandler#send(org.clazzes.remoting.cmd.LoadClassCmd)
     */
    public synchronized ClassBytes send(LoadClassCmd cmd) throws IOException {
        
        if (this.invocation != null || this.loadClass != null)
            throw new IOException("Concurrent remote invocations are not implemented.");
                  
        this.marshaler.writeObject(cmd);
        this.loadClass = cmd;
   
        try {
            
            this.wait();
            
        } catch (InterruptedException e) {
           
            this.loadClass = null;
            this.marshaler.close();
            throw new IOException("Wait for result of load class request has been interupted.");
        } 

        this.loadClass = null;
 
        if  (this.returnClassBytes != null) {
                
            ClassBytes ret = this.returnClassBytes.getClassBytes();
            this.returnClassBytes = null;
            return ret;
        }
        
        if (this.exception instanceof IOException) {
            
            Throwable e = this.exception;
            this.exception = null;
            throw (IOException)e;
        }

        if (this.exception != null) {
            IOException ioe = new IOException("Exception caught during read of response to remote invocation.");
            ioe.initCause(this.exception);
            this.exception = null;
            throw ioe;
        }
           
        throw new IllegalStateException("Illegal state while fetching reponse to load class request.");
    }
}
