/***********************************************************
 * $Id: RemotingClassLoader.java 143 2009-10-02 08:47:14Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 01.10.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.loading;

import java.io.IOException;

import org.clazzes.remoting.Destroyable;
import org.clazzes.remoting.RemoteInvoker;
import org.clazzes.remoting.helper.ClassHelper;

/**
 * This class loader allows to dynamically load classes from a
 * remote server. 
 *
 * This implementation does not take care for retrieving the compiled
 * class via {@link ClassLoader#getResourceAsStream(String)} by fetching
 * the resource with the extension <code>.class</code>.
 * 
 * This means that classes, which have been loaded from a remote server can not be
 * forwarded to another server again, because
 * {@link ClassHelper#getCompiledClass(String, ClassLoader)} works by fetching the
 * associated <code>.class</code> resource.
 * 
 * Implementing {@link ClassLoader#getResourceAsStream(String)} causes major headaches
 * and passing a remotely loaded class to another server seems to be a risk anyways.
 */
public class RemotingClassLoader extends ClassLoader implements Destroyable {

    RemoteInvoker client;
   
    /**
     * @param parent The parent class loader.
     * @param client The client used to load remote classes.
     */
    public RemotingClassLoader(ClassLoader parent, RemoteInvoker client) {
        super(parent);
        this.client = client;
    }

    private synchronized Class<?> fetchClass(String name) throws IOException {
        
        if (this.client == null)
            return null;
                
        if (!this.client.isConnected())
            this.client.connect();
        
        ClassBytes cb = this.client.loadClass(name);
      
        Class<?> clazz = this.defineClass(cb.getClassName(),cb.getClassBytes(),0,cb.getClassBytes().length);
        this.resolveClass(clazz);
            
        return clazz;
    }
    
    /* (non-Javadoc)
     * @see java.lang.ClassLoader#findClass(java.lang.String)
     */
    @Override
    protected Class<?> findClass(String name) throws ClassNotFoundException {
        
        try {
            Class<?> clazz = this.fetchClass(ClassHelper.getArrayComponentClass(name));
            
            if (clazz == null)
                throw new ClassNotFoundException("Error loading class ["+name+"]: client has been shut down."); 
            
            return clazz;
            
        } catch (IOException e) {
            
            // recover from transport errors by disconnecting.
            this.client.disconnect();
            
            throw new ClassNotFoundException("Error loading class ["+name+"] from remote server.",e); 
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.Destroyable#destroy()
     */
    public synchronized void destroy() {
        
        // stop loading classes from the remote server.
        
        if (this.client != null && this.client.isConnected())
            this.client.disconnect();
        
        this.client = null;
    }
}
