/***********************************************************
 * $Id: CompactMarshaler.java 147 2009-10-12 15:05:14Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 17.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.marshal.ObjectResolver;

/**
 * A marshaler, which marshals object in a field-by-field manner. It uses a compact
 * binary format on the underlying streams and uses a class registry, which might
 * contain a fixed set of classes to be marshaled.
 * 
 * The set of classes to be marshaled may be dynamically expanded, which is usually
 * not desirable, if a remoting server communicates with an implementation in a foreign
 * language like C++.
 * 
 * @author wglas
 */
public class CompactMarshaler extends MarshalerSupport {

    private static final Log log = LogFactory.getLog(CompactMarshaler.class);

    private final MarshalerClassRegistry classRegistry;
    private final ObjectResolver objectResolver;
    private final boolean extendClasses;
    private final boolean flattenCollections;
    
    private ClassLoader classLoader;
    private DataOutputStream dataOutputStream;
    private DataInputStream dataInputStream;
    
    private long lastWriteID;
    private Map<Object,Long> writeCache;
    private long lastWriteClassInfoID;
    private Map<ClassInfo,Long> writeClassInfoCache;

    private Map<Long,Object> prematureReadCache;
    private Map<Long,Object> readCache;
    private Map<Long,ClassInfo> readClassInfoCache;
    
    /**
     * @param objectResolver 
     * @param outputBufferSize
     * @param inputBufferSize
     */
    public CompactMarshaler(MarshalerClassRegistry classRegistry,
            boolean extendClasses,
            boolean flattenCollections,
            ObjectResolver objectResolver,
            int outputBufferSize, int inputBufferSize) {
        
        super(outputBufferSize, inputBufferSize);
        this.classRegistry = classRegistry;
        this.objectResolver = objectResolver;
        this.extendClasses = extendClasses;
        this.flattenCollections = flattenCollections;
    }

    /**
     * Flush the cache of read objects upon receiving {@link FieldMarshaler#OBJECT_FLUSH_STREAM_ID}.
     */
    void flushRead() {
        this.readCache.clear();
        this.prematureReadCache.clear();
    }
    
    /**
     * @param ref The referenced read object ID.
     * @return The cached object for the given object ID.
     */
    Object readCached(Long ref) {
        return this.readCache.get(ref);
    }
    
    /**
     * @param ref The referenced read object ID.
     * @return The cached, unresolved object for the given object ID.
     */
    Object readPrematureCached(Long ref) {
        return this.prematureReadCache.get(ref);
    }
    
    /**
     * Ache an object, which does not need to be resolved.
     * 
     * @param ref The read object ID. 
     * @param o The object to cache under the given ID.
     */
    void cacheRead(long ref, Object o) {
        if (log.isDebugEnabled())
            log.debug("Adding entry ["+ref+"/"+o.getClass().getName()+"] to read cache.");
        this.readCache.put(ref,o);
    }
    
    /**
     * Cache an object prior to resolving.
     * 
     * @param ref The read object ID. 
     * @param o The object to cache under the given ID.
     */
    void cacheReadPremature(long ref, Object o) {
        if (log.isDebugEnabled())
            log.debug("Adding entry ["+ref+"/"+o.getClass().getName()+"] to premature read cache.");
        this.prematureReadCache.put(ref,o);
    }
    
    /**
     * Cache an object after resolving. The unresolved object cached with
     * {@link #cacheReadPremature(long, Object)} is deleted from the
     * premature read cache.
     * 
     * @param ref The read object ID. 
     * @param o The object to cache under the given ID.
     */
    void cacheReadResolved(long ref, Object o) {
        
        this.prematureReadCache.remove(ref);
        
        if (log.isDebugEnabled())
            log.debug("Adding entry ["+ref+"/"+o.getClass().getName()+"] to resolve read cache.");
        this.readCache.put(ref,o);
    }
    
   /**
     * @param ref The referenced read class info ID.
     * @return The cached class info for the given object ID.
     */
    ClassInfo readClassInfoCached(Long ref) {
        return this.readClassInfoCache.get(ref);
    }
    
    /**
     * @param ref The read class info ID. 
     * @param o The class info to cache under the given ID.
     */
    void cacheClassInfoRead(long ref, ClassInfo ci) {
        this.readClassInfoCache.put(ref,ci);
    }
    
    /**
     * @param o The object to look up in the write cache.
     * @return The object ID of the given object, if the object has been written before.
     *         If <code>null</code> is returned, the object has not yet been written.
     */
    Long writeCached(Object o) {
        return this.writeCache.get(o);
    }

    /**
     * @param o The object to cache in the write cache for future references.
     * @return The newly generated object ID.
     */
    long cacheWrite(Object o) {
        
        long ref = ++this.lastWriteID;
        this.writeCache.put(o,ref);
        return ref;
    }
    
    /**
     * @param o The class info to look up in the write cache.
     * @return The ID of the given class info, if the class info has been written before.
     *         If <code>null</code> is returned, the class info has not yet been written.
     */
    Long writeClassInfoCached(ClassInfo ci) {
        return this.writeClassInfoCache.get(ci);
    }

    /**
     * @param o The class info to cache in the write cache for future references.
     * @return The newly generated class info ID.
     */
    long cacheClassInfoWrite(ClassInfo ci) {
        
        long ref = ++this.lastWriteClassInfoID;
        this.writeClassInfoCache.put(ci,ref);
        return ref;
    }
    
    /**
     * @param c The class to look up for writing.
     * @return The class info for writing.
     * @throws IllegalArgumentException
     * 
     * @see MarshalerClassRegistry#getClassInfo(Class, boolean)
     */
    ClassInfo getClassInfo(Class<?> c) throws IllegalArgumentException {
        return this.classRegistry.getClassInfo(c,this.extendClasses);
    }
    
    /**
     * @param c The class info to look up.
     * @return The class fields for reading/writing.
     * @throws IllegalArgumentException
     * 
     * @see MarshalerClassRegistry#getObjectMarshaler(ClassInfo, ClassLoader)
     */
     ObjectMarshaler getObjectMarshaler(ClassInfo ci) throws ClassNotFoundException {
        return this.classRegistry.getObjectMarshaler(ci,this.extendClasses ? this.classLoader : null);
    }
    
    /**
     * @return The data input stream set up in {@link #setup(OutputStream, InputStream, ClassLoader)}.
     */
    DataInputStream getDataInputStream() {
        return this.dataInputStream;
    }
    
    /**
     * @return The data output stream set up in {@link #setup(OutputStream, InputStream, ClassLoader)}.
     */
    DataOutputStream getDataOutputStream() {
        return this.dataOutputStream;
    }

    /**
     * @return The class loader used to load classes upon deserialization, if
     *         {@link CompactMarshaler#extendClasses} is set to <code>true</code>.
     */
    ClassLoader getClassLoader() {
        return this.classLoader;
    }
    
    /**
     * @return the object resolver used to intercept objects being read or written.
     */
    public ObjectResolver getObjectResolver() {
        return this.objectResolver;
    }

    /**
     * @return Whether the class registry should be auto-extended for new classes
     *         to be serialized.  
     */
    public boolean isExtendClasses() {
        return this.extendClasses;
    }

    /**
     * @return Whether to write collections in a specialized form.
     */
    public boolean isFlattenCollections() {
        return this.flattenCollections;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#close()
     */
    public void close() {
        
        try {
            if (this.dataInputStream != null)
                this.dataInputStream.close();
        } catch (IOException e) {
            log.warn("Error closing DataInputStream",e);
        }
        this.dataInputStream = null;
        
        try {
            if (this.dataOutputStream != null)
                this.dataOutputStream.close();
        } catch (IOException e) {
            log.warn("Error closing DataOutputStream",e);
        }
        this.dataOutputStream = null;
        this.readCache = null;
        this.writeCache = null;
        this.readClassInfoCache = null;
        this.writeClassInfoCache = null;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#readObject()
     */
    public Object readObject() throws ClassNotFoundException, IOException {
        
        if (this.dataInputStream == null)
            throw new EOFException("The marhaler has already been closed.");
        
        try {
            return ObjectFieldMarshaler.readObject(this);
            
        } catch (IllegalAccessException e) {
            
            IOException ioe = new IOException("Illegal field acces during deserialization.");
            ioe.initCause(e);
            throw ioe;
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#reset()
     */
    public void reset() throws IOException {
        
        if (log.isDebugEnabled())
            log.debug("Flushing write cache.");
        this.dataOutputStream.write(FieldMarshaler.OBJECT_FLUSH_STREAM_ID);
        this.dataOutputStream.flush();
        this.writeCache.clear();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#setup(java.io.OutputStream, java.io.InputStream, java.lang.ClassLoader)
     */
    public void setup(OutputStream os, InputStream is,
            ClassLoader inputClassLoader) throws IOException {

        try {
            this.dataOutputStream = new DataOutputStream(this.prepareOutputStream(os));
            this.dataInputStream = new DataInputStream(this.prepareInputStream(is));
            this.classLoader = inputClassLoader;
            
            this.lastWriteID = -1;
            this.writeCache = new IdentityHashMap<Object,Long>(256);
            this.lastWriteClassInfoID = -1;
            this.writeClassInfoCache = new HashMap<ClassInfo,Long>(64);
            
            this.prematureReadCache = new HashMap<Long,Object>(64);
            this.readCache = new HashMap<Long,Object>(256);
            this.readClassInfoCache = new HashMap<Long,ClassInfo>(64);
            
        } catch (IOException e) {
            
            is.close();
            
            if (this.dataOutputStream != null) {
                try {
                    this.dataOutputStream.close();
                } catch (Exception e2) {
                    log.warn("Error while closing output stream upon input stream setup problem",e2);
                }
                this.dataOutputStream = null;
            } else
                os.close();
            
            throw e;
        }
        
    }

    /**
     * This writes the object to the stream without flushing the write buffer.
     * This method might be used by reimplementations of CompactMarshaler which
     * obey a more complex flush policy.
     * 
     * @param obj The object to write
     * @throws IOException Upon errors.
     */
    protected void writeObjectNoFlush(Object obj) throws IOException {
        
        try {
            ObjectFieldMarshaler.writeObject(this,obj);

        } catch (ClassNotFoundException e) {

            IOException ioe = new IOException("Class not found during serialization.");
            ioe.initCause(e);
            throw ioe;
            
        } catch (IllegalArgumentException e) {

            IOException ioe = new IOException("Illegal argument during serialization.");
            ioe.initCause(e);
            throw ioe;
            
        } catch (IllegalAccessException e) {
        
            IOException ioe = new IOException("Illegal access during serialization.");
            ioe.initCause(e);
            throw ioe;

        }
    }
    
    /**
     * This method flushes the underlying stream without flushing the
     * write cache
     * 
     * This method might be used by reimplementations of CompactMarshaler which
     * obey a more complex flush policy.
     * 
     * @throws IOException Upon errors on the underlying stream.
     */
    protected void flushStream() throws IOException {
        this.dataOutputStream.flush();
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#writeObject(java.lang.Object)
     */
    public void writeObject(Object obj) throws IOException {
        
        this.writeObjectNoFlush(obj);
        this.reset();
    }
}
