/***********************************************************
 * $Id: FieldMarshaler.java 199 2010-09-06 14:33:40Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 17.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This is the base of the implementation of the compact marshaler.
 * 
 * Primitive types are written using the possibilities of {@link DataInputStream}
 * and {@link DataOutputStream} respectively. Hence integral values are written in
 * MSB byte order. (Most significant byte first).
 * 
 * @author wglas
 */
abstract class FieldMarshaler {

    private static final Log log = LogFactory.getLog(FieldMarshaler.class);
    
    /**
     * Stream format.
     * <pre>
     * +------ 0 -------+
     * | NULL_STREAM_ID |
     * +----------------+
     * </pre>
     */
    public static final int NULL_STREAM_ID               = 0x00;

    /**
     * Stream format.
     * <pre>
     * +-------- 0 --------+-- 1 --+
     * | BOOLEAN_STREAM_ID | {0|1} |
     * +-------------------+-------+
     * </pre>
     */
    public static final int BOOLEAN_STREAM_ID            = 0x01;

    /**
     * Stream format.
     * <pre>
     * +------- 0 ------+- 1 --+
     * | BYTE_STREAM_ID | byte |
     * +----------------+------+
     * </pre>
     */
    public static final int BYTE_STREAM_ID               = 0x02;

    /**
     * Stream format.
     * <pre>
     * +------- 0 -------+--- 1-2 ----+
     * | SHORT_STREAM_ID | short(MSB) |
     * +-----------------+------------+
     * </pre>
     */
    public static final int SHORT_STREAM_ID              = 0x03;

    /**
     * Stream format.
     * <pre>
     * +------ 0 ------+-- 1-4 ---+
     * | INT_STREAM_ID | int(MSB) |
     * +---------------+----------+
     * </pre>
     */
    public static final int INT_STREAM_ID                = 0x04;

    /**
     * Stream format.
     * <pre>
     * +------- 0 ------+--- 1-8 ---+
     * | LONG_STREAM_ID | long(MSB) |
     * +----------------+-----------+
     * </pre>
     */
    public static final int LONG_STREAM_ID               = 0x05;

    /**
     * Stream format.
     * <pre>
     * +------- 0 -------+------- 1-4 --------+
     * | FLOAT_STREAM_ID | float(ieee754/MSB) |
     * +-----------------+--------------------+
     * </pre>
     */
    public static final int FLOAT_STREAM_ID              = 0x06;

    /**
     * Stream format.
     * <pre>
     * +-------- 0 -------+-------- 1-8 --------+
     * | DOUBLE_STREAM_ID | double(ieee754/MSB) |
     * +------------------+---------------------+
     * </pre>
     */
    public static final int DOUBLE_STREAM_ID             = 0x07;

    /**
     * Stream format.
     * <pre>
     * +------- 0 ------+--- 1-2 ---+
     * | CHAR_STREAM_ID | char(MSB) |
     * +----------------+-----------+
     * </pre>
     */
    public static final int CHAR_STREAM_ID               = 0x08;

    /**
     * Stream format.
     * <pre>
     * +----------- 0 -----------+-------- 1-8 ---------+- 9-12 -+----- 13-(13+n) ------+
     * | BOOLEAN_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | b[0],b[1],...,b[n-1] |
     * +-------------------------+----------------------+--------+----------------------+
     * </pre>
     */
    public static final int BOOLEAN_ARRAY_STREAM_ID      = 0x09;

    /**
     * Stream format.
     * <pre>
     * +---------- 0 ---------+-------- 1-8 ---------+- 9-12 -+----- 13-(13+n) ------+
     * | BYTE_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | b[0],b[1],...,b[n-1] |
     * +----------------------+----------------------+--------+----------------------+
     * </pre>
     */
    public static final int BYTE_ARRAY_STREAM_ID         = 0x0a;

    /**
     * Stream format.
     * <pre>
     * +---------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+2*n) -----+
     * | SHORT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | s[0],s[1],...,s[n-1] |
     * +-----------------------+----------------------+--------+----------------------+
     * s[j]...MSB short
     * </pre>
     */
    public static final int SHORT_ARRAY_STREAM_ID        = 0x0b;
    
    /**
     * Stream format.
     * <pre>
     * +--------- 0 ---------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+4*n) -----+
     * | INT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | i[0],i[1],...,i[n-1] |
     * +---------------------+----------------------+--------+----------------------+
     * i[j]...MSB int
     * </pre>
     */
    public static final int INT_ARRAY_STREAM_ID          = 0x0c;
    
    /**
     * Stream format.
     * <pre>
     * +--------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+8*n) -----+
     * | LONG_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | l[0],l[1],...,l[n-1] |
     * +----------------------+----------------------+--------+----------------------+
     * l[j]...MSB long
     * </pre>
     */
    public static final int LONG_ARRAY_STREAM_ID         = 0x0d;
    
    /**
     * Stream format.
     * <pre>
     * +---------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+4*n) -----+
     * | FLOAT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | f[0],f[1],...,f[n-1] |
     * +-----------------------+----------------------+--------+----------------------+
     * f[j]...ieee754/MSB float
     * </pre>
     */
    public static final int FLOAT_ARRAY_STREAM_ID        = 0x0e;
    
    /**
     * Stream format.
     * <pre>
     * +---------- 0 -----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+8*n) -----+
     * | DOUBLE_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | d[0],d[1],...,d[n-1] |
     * +------------------------+----------------------+--------+----------------------+
     * d[j]...ieee754/MSB double
     * </pre>
     */
    public static final int DOUBLE_ARRAY_STREAM_ID       = 0x0f;

    /**
     * Stream format.
     * <pre>
     * +--------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+2*n) -----+
     * | CHAR_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | c[0],c[1],...,c[n-1] |
     * +----------------------+----------------------+--------+----------------------+
     * c[j]...MSB char
     * </pre>
     */
    public static final int CHAR_ARRAY_STREAM_ID         = 0x10;
    
    /**
     * Stream format.
     * <pre>
     * +-------- 0 -------+-------- 1-8 ---------+----- 9-10 ------+----- 11-(11+n) ------+
     * | STRING_STREAM_ID | object ID (MSB long) | n(unsigned MSB) | n modified UTF bytes |
     * +------------------+----------------------+-----------------+----------------------+
     * </pre>
     *
     * @see DataOutput#writeUTF(String)
     */
    public static final int STRING_STREAM_ID             = 0x11;
    
    /**
     * Object with full class header.
     * <pre>
     * +------- 0 --------+--------- 1-(n+3) ---------+------ (n+4)-(n+11) --------+ 
     * | OBJECT_STREAM_ID | writeUTF(class.getName()) | serialVerionUID (MSB long) |
     * +------------------+---------------------------+----------------------------+
     * 
     * +----- (n+12)-(n+19) ------+--- (n+20)-(n+27) ----+- (n+28)-(n+31) -+ 
     * | class info ID (MSB long) | object ID (MSB long) |    m(MSB int)   |
     * +--------------------------+----------------------+-----------------+
     * 
     * +---------+---------+----------+-----------+
     * | field 0 | field 2 |  ...     | field m-1 |
     * +---------+---------+----------+-----------+
     *
     * n...length of modified UTF bytes in class name.
     * m...number of fields written.
     * 
     * fields are written using their individual data representation. The order of fields
     * is according to the lexicographic ordering of the field names.
     * </pre>
     * 
     * @see DataOutput#writeUTF(String)
     */
    public static final int OBJECT_STREAM_ID             = 0x12;
      
    /**
     * Object with referenced class header.
     * <pre>
     * +--------- 0 -----------+---------- 1-8 -----------+-------- 9-16 --------+ 
     * | OBECT_CIREF_STREAM_ID | class info ID (MSB long) | object ID (MSB long) |
     * +-----------------------+--------------------------+----------------------+
     * 
     * +-- 17-20 ---+---------+---------+----------+-----------+
     * | m(MSB int) | field 0 | field 2 |  ...     | field m-1 |
     * +------------+---------+---------+----------+-----------+
     * 
     * m...number of fields written.
     * 
     * fields are written using their individual data representation with an individual tag.
     * The class info ID must have been previously serialized using
     * OBJECT_STREAM_ID or OBJECT_ARRAY_STREAM_ID.
     * </pre>
     */
    public static final int OBJECT_CIREF_STREAM_ID       = 0x13;
    
    /**
     * Reference to a previously serialized Object.
     * <pre>
     * +--------- 0 ----------+-------- 1-8 ---------+
     * | OBJECT_REF_STREAM_ID | object ID (MSB long) |
     * +----------------------+----------------------+
     *
     * The object ID must have been previously serialized using OBJECT_STREAM_ID,
     * OBJECT_CIREF_STREAM_ID, OBJECT_ARRAY_STREAM_ID, OBJECT_ARRAY_CIREF_STREAM_ID,
     * LIST_STREAM_ID, SET_STREAM_ID or MAP_STREAM_IP
     * </pre>
     */
    public static final int OBJECT_REF_STREAM_ID         = 0x14;
    
    /**
     * Object array with full component class header.
     * <pre>
     * +---------- 0 -----------+--------- 1-(n+3) ---------+------ (n+4)-(n+11) --------+ 
     * | OBJECT_ARRAY_STREAM_ID | writeUTF(class.getName()) | serialVerionUID (MSB long) |
     * +------------------------+---------------------------+----------------------------+
     * 
     * +----- (n+12)-(n+19) ------+--- (n+20)-(n+27) ----+- (n+28)-(n+31) -+ 
     * | class info ID (MSB long) | object ID (MSB long) |    m(MSB int)   |
     * +--------------------------+----------------------+-----------------+
     * 
     * +------+------+-------+--------+
     * | o[0] | o[1] |  ...  | o[m-1] |
     * +------+------+-------+--------+
     *
     * n...length of modified UTF bytes in component class name.
     * m...number of array members written.
     * 
     * members are written using their individual data representation with an individual tag.
     * The header class info is the array's component class.
     * </pre>
     * 
     * @see DataOutput#writeUTF(String)
     * @see Array#newInstance(Class, int)
     */
    public static final int OBJECT_ARRAY_STREAM_ID       = 0x15;
    
    /**
     * Object array with referenced component class header.
     * <pre>
     * +------------ 0 --------------+---------- 1-8 -----------+-------- 9-16 --------+ 
     * | OBECT_ARRAY_CIREF_STREAM_ID | class info ID (MSB long) | object ID (MSB long) |
     * +-----------------------------+--------------------------+----------------------+
     * 
     * +-- 17-20 ---+------+------+-------+--------+
     * | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
     * +------------+------+------+-------+--------+
     * 
     * m...number of array members written.
     * 
     * members are written using their individual data representation with an individual tag.
     * The class info ID of the array's component class must have been previously serialized
     * using OBJECT_STREAM_ID or OBJECT_ARRAY_STREAM_ID.
     * </pre>
     */
    public static final int OBJECT_ARRAY_CIREF_STREAM_ID = 0x16;
    
    /**
     * Stream format. Flushes the input object cache.
     * <pre>
     * +---------- 0 -----------+
     * | OBJECT_FLUSH_STREAM_ID |
     * +------------------------+
     * </pre>
     */
    public static final int OBJECT_FLUSH_STREAM_ID       = 0x17;
    
    /**
     * Optimized serialization of a list.
     * <pre>
     * +------ 0 -------+-------- 1-8 ---------+
     * | LIST_STREAM_ID | object ID (MSB long) |
     * +----------------+----------------------+
     *
     * +--- 9-12 ---+------+------+-------+--------+
     * | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
     * +------------+------+------+-------+--------+
     *
     * members are written using their individual data representation with an individual tag.
     * </pre>
     */
    public static final int LIST_STREAM_ID               = 0x18;
    
    /**
     * Optimized serialization of a set.
     * <pre>
     * +------ 0 ------+-------- 1-8 ---------+
     * | SET_STREAM_ID | object ID (MSB long) |
     * +---------------+----------------------+
     *
     * +--- 9-12 ---+------+------+-------+--------+
     * | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
     * +------------+------+------+-------+--------+
     *
     * members are written using their individual data representation with an individual tag.
     * </pre>
     */
    public static final int SET_STREAM_ID                = 0x19;
    
    /**
     * Optimized serialization of a map.
     * <pre>
     * +------ 0 ------+-------- 1-8 ---------+
     * | MAP_STREAM_ID | object ID (MSB long) |
     * +---------------+----------------------+
     *
     * +--- 9-12 ---+------+------+------+------+-------+--------+--------+
     * | m(MSB int) | k[0] | v[0] | k[1] | v[1] |  ...  | k[m-1] | v[m-1] |
     * +------------+------+------+------+------+-------+--------+--------+
     *
     * key/value pairs are written using their individual data representation with an individual tag.
     * </pre>
     */
    public static final int MAP_STREAM_ID                = 0x20;
    
    /**
     * Stream format for strings, which contain more than 65535 modified UTF-8 bytes.
     * Please note, that both the length of the string in chars and the length of the
     * serialized UTF bytes are written in order to allow an optimized deserialization
     * implementation which allocates minimal system resources.
     * <pre>
     * +-------- 0 -------+-------- 1-8 ---------+---- 9-12 -----+---- 13-16 ----+----- 17-(17+n) ------+
     * | STRING_STREAM_ID | object ID (MSB long) | string length | n(signed MSB) | n modified UTF bytes |
     * +------------------+----------------------+---------------+---------------+----------------------+
     * </pre>
     *
     * @see DataOutput#writeUTF(String)
     */
    public static final int LONG_STRING_STREAM_ID        = 0x21;

    private static final Map<Class<?>,Class<? extends FieldMarshaler>> impls;
    
    static {
        
        impls = new HashMap<Class<?>, Class<? extends FieldMarshaler>>(12);
        impls.put(boolean.class,BooleanFieldMarshaler.class);
        impls.put(byte.class,ByteFieldMarshaler.class);
        impls.put(short.class,ShortFieldMarshaler.class);
        impls.put(int.class,IntFieldMarshaler.class);
        impls.put(long.class,LongFieldMarshaler.class);
        impls.put(float.class,FloatFieldMarshaler.class);
        impls.put(double.class,DoubleFieldMarshaler.class);
        impls.put(char.class,CharFieldMarshaler.class);
    } 
    
    /**
     * Construct a marshaler for the given field out of the known implementations.
     * 
     * @param field A field of an object to be serialized.
     * @return A field marshaler dependent on the type of field.
     */
    public static FieldMarshaler newInstance(Field field) {
        
        Class<? extends FieldMarshaler> impl = impls.get(field.getType());
        
        if (impl == null) {
            
            if (StackTraceFieldMarshaler.THROWABLE_STACK_TRACE_FIELD.equals(field))
                return new StackTraceFieldMarshaler();
            else
                return new ObjectFieldMarshaler(field);
        }
        else {
            try {
                FieldMarshaler ret = impl.newInstance();
                ret.setField(field);
                return ret;
                
            } catch (Throwable e) {
               
                throw new RuntimeException("Cannot instantiate field marshaler of type ["+impl+"].",e);
            }
        }
    }
    
    protected Field field;
    
    /**
     * Default constructor, please call {@link #setField(Field)} after this constructor.
     */
    public FieldMarshaler() {
    }

    /**
     * @param field The field to be 
     */
    public FieldMarshaler(Field field) {
        super();
        this.setField(field);
    }
    
    /**
     * Initialize this marshaler for the given field.
     * 
     * @param field The field to marshal.
     */
    protected final void setField(Field field) {
        
        this.field = field;
        this.field.setAccessible(true);
    }
    
    /**
     * Read a tag an potentially flush the read object cache.
     * 
     * @param context The marshaler to read from.
     * @return The next non-flushing tag.
     * @throws IOException Upon read errors.
     */
    public static int readTag(final CompactMarshaler context) throws IOException {
        
        DataInputStream dis = context.getDataInputStream();
        
        int tag = dis.read();
        
        while (tag == OBJECT_FLUSH_STREAM_ID) {
            
            if (log.isDebugEnabled())
                log.debug("Flushing read cache.");
            
            context.flushRead();
            tag = dis.read();
        }
        
        if (tag == -1)
            throw new EOFException("Peer has closed socket during read of tag.");
        
        return tag;
    }
    
    /**
     * Write the handled field for the given marshaler.
     * 
     * @param context The marshaler, which holds all context information.
     * @param obj The object of which to write a field.
     * @throws IOException
     * @throws IllegalArgumentException
     * @throws IllegalAccessException
     * @throws ClassNotFoundException
     */
    public abstract void writeField(final CompactMarshaler context, final Object obj) throws IOException, IllegalArgumentException, IllegalAccessException, ClassNotFoundException;
    
    /**
     * Read the handled field from the given marshaler.
     * 
     * @param context The marshaler, which holds all context information.
     * @param obj The object of which to read a field.
     * @throws IllegalArgumentException
     * @throws IllegalAccessException
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public abstract void readField(final CompactMarshaler context, final Object obj) throws IllegalArgumentException, IllegalAccessException, IOException, ClassNotFoundException;

    /**
     * @return The field treated by this marshaler. 
     */
    public Field getField() {
        return this.field;
    }
}
