package org.clazzes.remoting.marshal.impl;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FilterInputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;

import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;

import org.clazzes.remoting.marshal.Marshaler;

public abstract class MarshalerSupport implements Marshaler {

    private int outputBufferSize;
    private int inputBufferSize;

    /**
     * Default constructor, which set the output and input buffer size to 0.
     */
    public MarshalerSupport() {
    }
    
    /**
     * Construct a JAVAMarshaler, with given input and output buffer sizes. 
     * @param outputBufferSize The size of the output buffer to use.
     * @param inputBufferSize The size of the input buffer to use.
     */
    public MarshalerSupport(int outputBufferSize, int inputBufferSize) {
        super();
        this.outputBufferSize = outputBufferSize;
        this.inputBufferSize = inputBufferSize;
    }

    /**
     * Prepare the transport output stream for use with ObjectOutputStream.
     * This may involve wrapping the stream with a {@link BufferedInputStream} or
     * some other sort of {@link FilterInputStream} like {@link CipherInputStream} or
     * {@link GZIPOutputStream}.
     * 
     * The default implementation wraps the output stream with {@link BufferedInputStream},
     * if the output buffer size is set to a positive value.
     * 
     * @param os The raw OutputStream as passed to {@link #setup(OutputStream, InputStream, ClassLoader)}.
     * @return A wrapped output stream.
     * @throws IOException Upon errors setting up the stream.
     */
    protected OutputStream prepareOutputStream(OutputStream os) throws IOException {
        if (this.outputBufferSize <= 0)
            return os;
        else
            return new BufferedOutputStream(os,this.outputBufferSize);
    }

    /**
     * Prepare the transport input stream for use with ObjectInputStream.
     * This may involve wrapping the stream with a {@link BufferedOutputStream} or
     * some other sort of {@link FilterOutputStream} like {@link CipherOutputStream} or
     * {@link GZIPOutputStream}.
     * 
     * The default implementation wraps the input stream with {@link BufferedOutputStream},
     * if the input buffer size is set to a positive value.
     * 
     * 
     * @param is The raw InputStream as passed to {@link #setup(OutputStream, InputStream, ClassLoader)}.
     * @return A wrapped InputStream.
     */
    protected InputStream prepareInputStream(InputStream is) throws IOException {
        if (this.inputBufferSize <= 0)
            return is;
        else
            return new BufferedInputStream(is,this.inputBufferSize);
    }

    /**
     * @return The buffer size to use for the {@link BufferedOutputStream} used.
     *         If set to 0, no output buffering is performed.
     */
    public int getOutputBufferSize() {
        return this.outputBufferSize;
    }

    /**
     * @param outputBufferSize The output buffer size to set.
     */
    public void setOutputBufferSize(int outputBufferSize) {
        this.outputBufferSize = outputBufferSize;
    }

    /**
     * @return The buffer size to use for the {@link BufferedInputStream} used.
     *         If set to 0, no input buffering is performed.
     */
    public int getInputBufferSize() {
        return this.inputBufferSize;
    }

    /**
     * @param inputBufferSize The input buffer size to set.
     */
    public void setInputBufferSize(int inputBufferSize) {
        this.inputBufferSize = inputBufferSize;
    }

}