/***********************************************************
 * $Id: ServerConnectionHandler.java 139 2009-10-01 14:16:18Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 16.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.server;

import java.io.Closeable;
import java.io.EOFException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.SocketException;
import java.rmi.server.UID;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Broadcastable;
import org.clazzes.remoting.CallbackInvocationContext;
import org.clazzes.remoting.InvalidTargetException;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.cmd.CallBackCmd;
import org.clazzes.remoting.cmd.Cmd;
import org.clazzes.remoting.cmd.CmdVisitorSupport;
import org.clazzes.remoting.cmd.ExchangeClientUIDCmd;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.cmd.InvocationExceptionCmd;
import org.clazzes.remoting.cmd.InvocationResultCmd;
import org.clazzes.remoting.cmd.LoadClassCmd;
import org.clazzes.remoting.cmd.ReturnClassBytesCmd;
import org.clazzes.remoting.helper.ClassHelper;
import org.clazzes.remoting.loading.ClassBytes;
import org.clazzes.remoting.marshal.Marshaler;

/**
 * A runnable connection handler for a single server connection of a client.
 * 
 * @author wglas
 */
public class ServerConnectionHandler extends CmdVisitorSupport implements CallbackInvocationContext,
        Runnable, Closeable {

    private static final Log log = LogFactory.getLog(ServerConnectionHandler.class);
    
    private final ServerRegistry serverRegistry;
    private final UID clientUID;
    private final Marshaler marshaler;
    private final Map<String,Object> meta;
    
    private boolean invocationRunning;
    private CallBackCmd invocation;
    private InvocationResultCmd invocationResult;
    private InvocationExceptionCmd invocationException;
    private Throwable exception;
    
    private Map<String, Object> buildMetaData(Map<String, Object> m) {
        
        if (m == null)
            return this.meta;
        
        if (this.meta != null)
            m.putAll(this.meta);
        
        return m;
    }

    /**
     * @param serverRegistry 
     * @param marshaler The marshaler used for encapsulating the socket.
     * @param clientUID The generated UID of the client.
     * @param meta 
     */
    public ServerConnectionHandler(ServerRegistry serverRegistry, UID clientUID,
                                   Marshaler marshaler, Map<String, Object> meta) {
        super();
        this.serverRegistry = serverRegistry;
        this.clientUID = clientUID;
        this.marshaler = marshaler;
        this.meta = meta;
        this.invocationRunning = false;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.CallbackInvocationContext#getBroadcastable()
     */
    public Broadcastable getBroadcastable() {
        
        return this.serverRegistry;
    }

    private class CallBackCmdVisitor extends CmdVisitorSupport {

        @Override
        public void visit(InvocationExceptionCmd cmd) {
            
            synchronized(ServerConnectionHandler.this) {
                
                if (ServerConnectionHandler.this.invocation == null || ServerConnectionHandler.this.invocationResult != null || ServerConnectionHandler.this.invocationException != null) {
                    super.visit(cmd);
                    return;
                }
                
                ServerConnectionHandler.this.invocationException = cmd;
            }
       }

        @Override
        public void visit(InvocationResultCmd cmd) {
            
            synchronized(ServerConnectionHandler.this) {
                
                if (ServerConnectionHandler.this.invocation == null || ServerConnectionHandler.this.invocationResult != null || ServerConnectionHandler.this.invocationException != null) {
                    super.visit(cmd);
                    return;
                }
                
                ServerConnectionHandler.this.invocationResult = cmd;
            }
         }
        
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.CallbackInvocationContext#callBack(java.rmi.server.UID, java.lang.Object, java.util.Map)
     */
    public Object callBack(UID callbackUID, Object arg,
            Map<String, Object> metadata) throws InvocationTargetException,
            IOException {
        
        CallBackCmd invocation = new CallBackCmd(callbackUID,arg,metadata);
        
        boolean synchronousOperation;
        
        synchronized (this) {
            
            if (this.invocation != null)
                throw new IOException("Concurrent remote invocations are not implemented.");
              
            this.marshaler.writeObject(invocation);
            this.invocation = invocation;
            synchronousOperation = this.invocationRunning;
        }
        
        if (synchronousOperation) {
        
            Object obj;
            try {
            
                obj = this.marshaler.readObject();
                Cmd.acceptCmd(obj,new CallBackCmdVisitor());

            } catch (Throwable e) {
                
                synchronized (this) { this.invocation=null; this.marshaler.close(); }
                
                IOException ioe = new IOException("Exception caught during read of reponse to callback invocation.");
                ioe.initCause(this.exception);
                throw ioe;
            }
            
            
        } else {
            try {
            
                synchronized(this.invocation) {
                    this.invocation.wait();
                }
            } catch (InterruptedException e) {
                
                synchronized (this) { this.invocation=null; this.marshaler.close(); }
                
                throw new IOException("Wait for invocation result has been interrupted.");
            }
        }
        
        synchronized (this) {
            
            this.invocation = null;
        
            if  (this.invocationResult != null) {
                
                Object ret = this.invocationResult.getResult();
                this.invocationResult = null;
                return ret;
            }
            if  (this.invocationException != null) {
                
                Throwable e = this.invocationException.getException();
                this.invocationException = null;
                throw new InvocationTargetException(e);
            }
            
            // transport error from read.
            if (this.exception instanceof IOException) {
                throw (IOException)this.exception;
            }
            
            // RuntimeException from read.
            if (this.exception != null) {
                IOException ioe = new IOException("Exception caught during read of reponse to callback invocation.");
                ioe.initCause(this.exception);
                throw ioe;
            }
     
            throw new IllegalStateException("Illegal state while fetching reponse to callback invocation.");
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationContext#getClientUID()
     */
    public UID getClientUID() {
        
        return this.clientUID;
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {
        
        if (log.isDebugEnabled())
            log.debug("Connection handler for client ["+this.getClientUID()+"] has been started.");
           
        try {
            
            synchronized (this) {
                   
                this.marshaler.writeObject(new ExchangeClientUIDCmd(this.getClientUID()));
            }
     
            
            while (true) {

                Object obj = this.marshaler.readObject();
                    
                Cmd.acceptCmd(obj,this);
            }
                
        } catch (Throwable e) {
                
           if (e instanceof EOFException) {
                
                if (log.isTraceEnabled())
                    log.trace("EOF while receiving an object from client ["+this.getClientUID()+"]",e);
           }
           else if (e instanceof SocketException) {
               
               if (log.isTraceEnabled())
                   log.trace("Socket closed while receiving an object from client ["+this.getClientUID()+"]",e);
           }
           else
                log.error("Error while receiving an object from client ["+this.getClientUID()+"]",e);
                
           synchronized(this) {
                    
                if (this.invocation != null) {
                    
                    this.exception = e;
                    
                    synchronized(this.invocation) {
                        this.invocation.notify();
                    }
                }
                this.marshaler.close();
            }
            this.serverRegistry.clientTerminated(this.clientUID);
        }
    }

    /**
     * Close the server connection. Please mind that the connection handler calls
     * {@link ServerRegistry#clientTerminated(UID)} asynchronously.
     */
    public synchronized void close() throws IOException {
       
        this.marshaler.close();
    }
    
    /**
     * @param cmd The invocation command to send to the client represented by this
     *            connection.
     * @throws IOException Upon errors from the marshaler.
     */
    public synchronized void broadcast(InvocationCmd cmd) throws IOException {
        
        this.marshaler.writeObject(cmd);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.cmd.CmdVisitorSupport#visit(org.clazzes.remoting.cmd.InvocationCmd)
     */
    public void visit(InvocationCmd cmd) {
        String subSystem = cmd.getSubSystem();
        
        InvocationHandler handler = this.serverRegistry.getInvocationHandler(subSystem);
        
        Cmd response;
        
        if (handler == null) {
            
          response=
              new InvocationExceptionCmd(
                      new InvalidTargetException("Subsystem ["+subSystem+"] does not exist."));
        
        } else {
            
            try {
                
                synchronized(this) {
                    this.invocationRunning = true;
                }
                    
                Object res = handler.invoke(this,cmd.getArg(),this.buildMetaData(cmd.getMetadata()));
                
                response = new InvocationResultCmd(res);
                
            } catch (Throwable e) {
                
                if (log.isDebugEnabled())
                    log.debug("Invocation threw excpetion",e);
                    
                response = new InvocationExceptionCmd(e);
            }
        }
        
        synchronized(this) {
            this.invocationRunning = false;
            
            try {
                this.marshaler.writeObject(response);

            } catch (IOException e) {
                
                throw new RuntimeException("I/O exception writing invocation response.",e);
            }
        }
        
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.cmd.CmdVisitorSupport#visit(org.clazzes.remoting.cmd.LoadClassCmd)
     */
    @Override
    public void visit(LoadClassCmd cmd) {
        
        String className = cmd.getClassName();
        
        if (log.isDebugEnabled())
            log.debug("Received request to load class ["+className+"].");
        
        ClassBytes classBytes = null;
        
        byte[] bytes;
        try {
            bytes = ClassHelper.getCompiledClass(className,this.serverRegistry.getMarshalClassLoader());
        } catch (Exception e) {
            throw new RuntimeException("Exception while getting compiled class ["+className+"].",e);
        }
        
        if (bytes != null)
            classBytes = new ClassBytes(className,bytes);
        
        if (log.isDebugEnabled())
            log.debug("Writing classBytes ["+classBytes+"].");

        ReturnClassBytesCmd response = new ReturnClassBytesCmd(classBytes);
        
        synchronized(this) {
        
            try {
                this.marshaler.writeObject(response);

            } catch (IOException e) {
                throw new RuntimeException("I/O exception writing load class response.",e);
            }
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.cmd.CmdVisitorSupport#visit(org.clazzes.remoting.cmd.InvocationResultCmd)
     */
    public void visit(InvocationResultCmd cmd) {
        
        synchronized(this) {
            
            if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
                super.visit(cmd);
                return;
            }
            
            this.invocationResult = cmd;
        }
        
        synchronized(this.invocation) {
            this.invocation.notify();
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.cmd.CmdVisitorSupport#visit(org.clazzes.remoting.cmd.InvocationExceptionCmd)
     */
    public void visit(InvocationExceptionCmd cmd) {
        
        synchronized(this) {
            
            if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
                super.visit(cmd);
                return;
            }
            
            this.invocationException = cmd;
        }
        
        synchronized(this.invocation) {
            this.invocation.notify();
        }        
    }

}
